/*
 * Copyright 2001-2002, 2010 pilot <pilot@monkey.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/time.h>

#include "arirang.h"
#include "version.h"

char *html = NULL;
FILE *htmlreport;
int wflags = 0;
int fflags = 0;
int ftype = 0;
int sslflags = 0;
short int CONNECT_TIMEOUT = 5;
int scan_count = 0;
int rule_grab = 0;
int rflags = 0;
int ruby = 0;
char *ariruby = NULL;
FILE *rubyfd;

int
main(int argc, char *argv[])
{
    int pcount = 0;

    unsigned short int MAX_PROCESSES = 30;
    unsigned short int RULE_PROCESSES = 0;
    unsigned short int IPRANGE_PROCESSES = 30;

    unsigned int http_port = 80;
    unsigned long start = 0, end = 0, count;
    unsigned short int statuscode = 1;

    int ch;
    char *server = NULL;
    char *filename = NULL;
    FILE *scanlist, *tmp_iprange;
    char list[50];

    char *rulefile = NULL;
    int gflags = 0, oflags = 0, hflags = 0,
        mflags = 0;

    struct hostent *host_cidr;
    struct in_addr host_addr;
    int cidr_domain = 0;

    char *slash = NULL;
    unsigned long cidr =0, mask=0;
    int bits = 0;

    unsigned long fcount = 0;
    char *fip;
    int fnum = 0;
    
    unsigned long sec;
    unsigned long usec;
    struct timeval tv1, tv2;
    gettimeofday(&tv1, 0);

    if (argc < 3)
        usage();

    scrview();
    opterr = 0;

    while ((ch = getopt(argc, argv, "St:w:R:GOp:f:h:s:e:r:vP:")) != -1) {
        switch (ch) {
		case 'R':
			ruby = 1;
			ariruby = optarg;
			gflags = 1;
			break;
        case 'S':
            sslflags = 1;
            http_port = 443;
            break;
        case 't':
            CONNECT_TIMEOUT = atoi(optarg);
            break;
        case 'w':
            wflags = 1; 
            html = optarg;
            break;
        case 'G':
            gflags = 1;
            break;
        case 'O':
            oflags = 1;
            break;
        case 'p':
            http_port = atoi(optarg);
            break;
        case 'f':
            fflags = 1;
            filename = optarg;
            break;
        case 'h':
            hflags = 1;
            server = optarg;
            break;
        case 's':
            start = inet_addr(optarg);
            break;
        case 'e':
            end = inet_addr(optarg);
            break;
        case 'r':
            rflags = 1;
            rulefile = optarg;
            break;
        case 'v':
            statuscode = 2;
            break;
        case 'P':
            MAX_PROCESSES = atoi(optarg);
            RULE_PROCESSES = atoi(optarg);
            IPRANGE_PROCESSES = atoi(optarg);
            break;
        default:
            usage();
        }
    }

#ifdef DEBUG
    printf("CONNECT_TIMEOUT: %d \n", CONNECT_TIMEOUT);
    printf("wflags: %d \n", wflags);
    printf("sslflags: %d \n", sslflags);
    printf("gflags: %d \n", gflags);
    printf("oflags: %d \n", oflags);
    printf("ports: %d \n", http_port);
    printf("fflags: %d \n", fflags);
    printf("ruby: %d \n", ruby);
    printf("ariruby: %s \n", ariruby);
#endif

    if (hflags) {
        if ((slash = strchr(server, '/')) != NULL)
            mflags = 1;
    }   

    if (start && end) {
        if (start > end) {
            fprintf(stdout, "start end address incorrect \n\n");
            exit(0);
        }
    }
   
    if (sslflags) 
        fprintf(stdout, "starting arirang %s + SSL    %s \n\n", VERSION, HOMEPAGE);
    else if (ruby)
		fprintf(stdout, "starting arirang %s + Ruby    %s \n\n", VERSION, HOMEPAGE);
	else	
        fprintf(stdout, "starting arirang %s    %s \n\n", VERSION, HOMEPAGE);

    if (rflags)
        fprintf(stdout, "%s - %s\n\n", SCANRULE_MSG, rulefile);

	if ((ruby && wflags) || (ruby && rflags)) {
    	fprintf(stderr, "arirang ruby script doesn't support with -w or -r \n");
        exit(1);
    }

    if (mflags) {
        *slash++ = '\0'; 

/* check domain or ip */
        fip = server;
        while (*fip != '\0') {
        if ((*fip >= 'a' && *fip <= 'z')  || (*fip >= 'A' && *fip <= 'Z'))
            fnum = 1;
         fip++;
        }
        if (fnum == 1) {
            if ((host_cidr = gethostbyname(server)) == NULL) {
                fprintf(stderr, "host not found \n\n");
                exit(0);
            }
        host_addr.s_addr  = *((unsigned long* ) host_cidr->h_addr_list[0]);
        fprintf(stdout, "%s - %s \n", server, inet_ntoa(host_addr)); 

        server = inet_ntoa(host_addr);
        cidr_domain = 1;
        }

        cidr = inet_addr(server);

        if (fnum != 1 || cidr_domain) {
        bits = atoi(slash);

        if (bits > 0 && bits < 33) { 
            mask = 0xFFFFFFFF << (32 - bits);
#ifdef DEBUG
            printf("bits: %d,  mask: %lu\n", bits, mask);
#endif
            cidr = ntohl(cidr);
            start = htonl(cidr & mask);
            end = htonl(cidr | ~mask);
        }
        else {
            fprintf(stdout, "bits incorrect \n\n");
            exit(0);
        }
      }
    fnum = 0;
    }

    if (wflags) {
      if ((htmlreport = fopen(html, "w+")) == NULL) {
          fprintf(stderr, "File %s write error\n", html);
          exit(1);
      }
      fprintf(stdout, "arirang scan report file - %s \n\n", html);
      htmlhead(html);
    }

	if (ruby) {
		if ((rubyfd = fopen(ariruby, "rt")) == NULL) {
			fprintf(stderr, "arirang ruby script - %s read error \n", ariruby);
			exit(1);
		}
		fprintf(stdout, "arirang ruby script file - %s \n\n", ariruby);
	}


    if (gflags && hflags && mflags == 0) {
        scan_count++;
        grabhead(server, http_port, 0);
    }

    if (oflags && hflags && mflags == 0) {
        scan_count++;
        grabhead(server, http_port, 1);
    }

    if (fflags == 1 && rflags == 0 && mflags == 0) {

        if ((scanlist = fopen(filename, "rt")) == NULL) {
            fprintf(stderr, "File %s open error\n", filename);
            exit(0);
        }

/* -f with process */
        pcount = 0;
        while (!feof(scanlist)) {
            fscanf(scanlist, "%s\n", list);

/* check domain or ip */
        fip = list;
        while (*fip != '\0') {
        if ((*fip >= 'a' && *fip <= 'z')  || (*fip >= 'A' && *fip <= 'Z'))
            fnum = 1;
         fip++;
        }

        if (fnum == 0) {
            fcount =  inet_addr(list);
            fcount = ntohl(fcount);
                                                                                          
            if ((fcount & 0xff) == 0 || (fcount & 0xff) == 255) 
                continue;
        }
        fnum = 0;

            if (pcount > MAX_PROCESSES)
                wait(NULL);

            switch (fork()) {
            case 0:
#ifdef DEBUG
            printf("%s \n", list);
#endif

            if (gflags) {
                grabhead(list, http_port, 0);
            }
            if (oflags)
                grabhead(list, http_port, 1);

                exit(0);
            case -1:
                perror("fork err");
                exit(-1);
            default:
                pcount++;
                break;
            }
        }
        fclose(scanlist);

        while (pcount--) {
            scan_count++;
            wait(NULL);
        }
    }

    if ((start && end && rflags == 0) || (mflags == 1 && rflags == 0)) {
        pcount = 0;

        for (count = ntohl(start); count <= ntohl(end); count++) {
            if ((count & 0xff) == 0 || (count & 0xff) == 255)
                continue;
            if (pcount > MAX_PROCESSES)
                wait(NULL);

            switch (fork()) {
            case 0:
                if (oflags)
                    grabheadW(count, http_port, 0, 1);
                else
                    grabheadW(count, http_port, 0, 0);
                
                exit(0);
            case -1:
                perror("fork");
                exit(-1);
            default:
                pcount++;
                break;
            }
        }
        while (pcount--) {
            scan_count++;   
            wait(NULL);
        }
    }
    if (rflags && statuscode && hflags && mflags == 0) {
        if (RULE_PROCESSES == 0)
            scanrule(server, http_port, rulefile, statuscode);
        if (RULE_PROCESSES != 0)
            scanruleP(server, http_port, rulefile, statuscode, RULE_PROCESSES);
    }
/* -f with rule */
    if (rflags && statuscode && fflags && mflags == 0) {
        ftype = 1;
        if ((tmp_iprange = fopen(USED_IP_RANGE, "w")) == NULL) {
            fprintf(stderr, "File %s write error", USED_IP_RANGE);
            exit(1);
        }
        fclose(tmp_iprange);
        fprintf(stdout, "please wait - arirang make %s file for network scan\n\n", USED_IP_RANGE);

        if ((scanlist = fopen(filename, "rt")) == NULL) {
            fprintf(stderr, "File %s open error\n", filename);
            exit(0);
        }

        pcount = 0;
        while (!feof(scanlist)) {
            fscanf(scanlist, "%s\n", list);

/* check domain or ip */
        fip = list;
        while (*fip != '\0') {
        if ((*fip >= 'a' && *fip <= 'z')  || (*fip >= 'A' && *fip <= 'Z'))
            fnum = 1;
        fip++;
        }

        if (fnum == 0) {
            fcount =  inet_addr(list);
            fcount = ntohl(fcount);

            if ((fcount & 0xff) == 0 || (fcount & 0xff) == 255)
                continue;
         }
        fnum = 0;

            if (pcount > MAX_PROCESSES)
                wait(NULL);

            switch (fork()) {
            case 0:
        
                if (gflags)
                    grabhead(list, http_port, 0);
                if (oflags)
                    grabhead(list, http_port, 1);

                exit(0);
            case -1:
                perror("fork err");
                exit(-1);
            default:
                pcount++;
                break;
            }
        }
        fclose(scanlist);

        while (pcount--) {
            scan_count++;
            wait(NULL);
        }

        ftype = 0;

        if ((scanlist = fopen(USED_IP_RANGE, "rt")) == NULL) {
            fprintf(stderr, "File %s open error\n", USED_IP_RANGE);
            exit(1);
        }
        while (!feof(scanlist)) {
            fscanf(scanlist, "%s\n", list);
            rule_grab = 1;
            if (oflags)
                grabhead(list, http_port, 1);
            else
                grabhead(list, http_port, 0);

			rule_grab = 0;
            scanruleP(list, http_port, rulefile, statuscode, IPRANGE_PROCESSES); 
        }
        fclose(scanlist);
    }

    if ((rflags && start && end) || (rflags && mflags == 1)) {
        if ((tmp_iprange = fopen(USED_IP_RANGE, "w")) == NULL) {
            fprintf(stderr, "File %s write error", USED_IP_RANGE);
            exit(1);
        }
        fclose(tmp_iprange);
        fprintf(stdout, "please wait - arirang make %s file for network scan\n\n", USED_IP_RANGE);

        pcount = 0;

        for (count = ntohl(start); count <= ntohl(end); count++) {
            if ((count & 0xff) == 0 || (count & 0xff) == 255)
                continue;
            if (pcount > IPRANGE_PROCESSES)
                wait(NULL);

            switch (fork()) {
            case 0:
                if (oflags)
                    grabheadW(count, http_port, 1, 1);
                else
                    grabheadW(count, http_port, 1, 0);
                exit(0);
            case -1:
                perror("fork");
                exit(-1);
            default:
                pcount++;
                break;
            }
        }
        while (pcount--) {
            scan_count++;
            wait(NULL);
        }

        if ((scanlist = fopen(USED_IP_RANGE, "rt")) == NULL) {
            fprintf(stderr, "File %s open error\n", USED_IP_RANGE);
            exit(1);
        }
        while (!feof(scanlist)) {
            fscanf(scanlist, "%s\n", list);
            rule_grab = 1;
            if (oflags)
                grabhead(list, http_port, 1);
            else
                grabhead(list, http_port, 0);

			rule_grab = 0;
            scanruleP(list, http_port, rulefile, statuscode, IPRANGE_PROCESSES);
        }
        fclose(scanlist);

    }

    gettimeofday(&tv2, 0);

    if (tv2.tv_usec <= tv1.tv_usec) { 
        tv2.tv_sec--;
        tv2.tv_usec = tv2.tv_usec + 1000000;
    } 

    sec = tv2.tv_sec - tv1.tv_sec;
    usec = (tv2.tv_usec - tv1.tv_usec) / 1000; 
    if (scan_count == 1)
        fprintf(stdout, "\narirang scan result: %d host %ld.%ld seconds\n\n", scan_count, sec, usec);
    else
        fprintf(stdout, "\narirang scan result: %d hosts %ld.%ld seconds\n\n", scan_count, sec, usec);

    return 0;
}
