/*
 * Copyright 2001-2002, 2010 pilot <pilot@monkey.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>

#include <ruby.h>
#include "arirang.h"

extern char *html;
extern FILE *htmlreport;
extern int wflags;
extern int fflags;
extern int ftype;
extern int sslflags;
extern short int CONNECT_TIMEOUT;
extern int scan_count;
extern int rflags;
extern int rule_grab;
int ralarm = 0;

extern int ruby;
extern char *ariruby;
extern FILE *rubyfd;

void
timegrab(int sig)
{
/*    int sock;
    close(sock); */
    exit(0);
}

void
connout(int sig)
{
/* rule scan set = 1 without grab exit*/
    if (rflags && rule_grab) 
    {
        ralarm = 1;
    }
    else {
/*        fprintf(stderr, "connect timeout \n"); */
        exit(0);
    }
}


int
recvtimeout(int s, char *buf, int len, int timeout)
{
    fd_set  fds;
    int n;
    struct timeval  tv;

    FD_ZERO(&fds);
    FD_SET(s, &fds);
    tv.tv_sec = timeout;
    tv.tv_usec = 0;

    n = select(s + 1, &fds, NULL, NULL, &tv);
    if (n == 0)
        return -2;
    if (n == -1)
        return -1;

    return recv(s, buf, len, 0);
}

void
grabhead(char *scanhost, unsigned int port, short int option)
{
    int sock;
    int nb = 0;
    struct sockaddr_in http;
    struct hostent *he;

//    char osfinger[32];
    char *osfinger = NULL;
    char headbuf[1024];

    char *ptr = NULL, *ptr2, *buf_ptr;
    FILE *tmp_iprange;

    SSL_CTX *ctx = NULL;
    SSL *ssl = NULL;
    int err;
    int sread = 0;

    char *rbuf;

    VALUE ari_send;
    VALUE ari_recv;
    VALUE ari_msg;
	VALUE ari_host;
    VALUE rstr;
    VALUE rhost;
    
    if (wflags) {
        if ((htmlreport = fopen(html, "a+")) == NULL) {
            fprintf(stderr, "File %s write error\n", html);
            exit(1);
        }
    }

    if ((he = gethostbyname(scanhost)) == NULL) {
        perror("host not found");
        /* close(sock);  */
        exit(0);
    }
    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
        perror("socket");
        close(sock);
        exit(0);
    }
    http.sin_family = AF_INET;
    http.sin_port = htons(port);
    http.sin_addr = *((struct in_addr *) he->h_addr);
    bzero(&(http.sin_zero), 8);

    signal(SIGALRM, connout);
    alarm(2);

    if (connect(sock, (struct sockaddr *) & http, sizeof(struct sockaddr)) == -1) {
    /*  rule scan continue without grabhead exit */ 
        if (rflags && rule_grab) {
            ralarm = 0; 
            close(sock);
            return;    
        }
        close(sock);
        exit(0);
    }

    /* rule scan continue without grabhead exit */
    if (ralarm) {
        ralarm = 0;
        close(sock);
        return;  
    }

    alarm(0);

    if (sslflags) {
        ctx = ssl_init_ctx();
        ssl = ssl_init_ssl(ctx, sock);
    }


    if (ruby) {
        ruby_init();
        ruby_init_loadpath();

        rb_define_variable("$ari_send", &ari_send);

        rb_load_file(ariruby);
        ruby_exec();  
        rbuf = StringValuePtr(ari_send);

		if (strlen(rbuf) == 0) {
			fprintf(stderr, "please set a data in $ari_send variable \n");
			exit(0);
		}

#ifdef DEBUG
        printf("ari_send size = %d \n", strlen(rbuf));
#endif
        if ((osfinger = (char *)malloc(strlen(rbuf) + 1)) == NULL)  {
            perror("malloc failed");
            close(sock);
            exit(0);
        }
        strcpy(osfinger, rbuf);
    } 
    else if (option) {
        if ((osfinger = (char *)malloc(strlen(SEND_OPTION) + 1)) == NULL) {
            perror("malloc failed");
            close(sock);
            exit(0);
        }
        strcpy(osfinger, SEND_OPTION);
    }
    else {
        if ((osfinger = (char *)malloc(strlen(SEND_URI) + 1)) == NULL) {
            perror("malloc failed");
            close(sock);
            exit(0);
        }
        strcpy(osfinger, SEND_URI);
    }

/*  printf("osfinger = %s \n", osfinger); */

    if (sslflags) {
        err = SSL_write(ssl, osfinger, strlen(osfinger)); 
        if (err <= 0) {
        close(sock);
        fprintf(stdout, "%s - ssl write time out \n", scanhost);
        exit(0);
        }

    ssl_error(ssl, err, scanhost);
#ifdef DEBUG
     printf("error = %d\n", err); 
#endif
     sread = ssl_read_timeout(sock, ssl, headbuf, sizeof(headbuf) -1, CONNECT_TIMEOUT);

     if (ruby) {
         rstr = rb_str_new2(headbuf);
         rb_load_file(ariruby);
         rb_define_variable("$ari_recv", &rstr);
    }

        if (sread <= 0 ) {
            close(sock);
        if (wflags)
            fprintf(htmlreport, "<font size=\"2\"><a href=\"https://%s\">%s</a> - ssl read time out<br></font>", 
                scanhost, scanhost); 
        else
            fprintf(stdout, "%s - ssl read time out \n", scanhost); 
        exit(0);
        }
#ifdef DEBUG
     printf("sread  = %d \n", sread); 
#endif
    }
    else {
    send(sock, osfinger, strlen(osfinger), 0);
    nb = recvtimeout(sock, headbuf, sizeof(headbuf), CONNECT_TIMEOUT);
        if (ruby) {
            rstr = rb_str_new2(headbuf);
 	        rb_load_file(ariruby); 
            rb_define_variable("$ari_recv", &rstr);
        }

    }

    if (osfinger != NULL)
        free(osfinger);

    if (nb == -1 || nb == -2 ) {
        if (wflags)
            fprintf(htmlreport, "<font size=\"2\">%s - maybe unknown network device or filter by firewall<br></font>", scanhost);
        else
            fprintf(stdout, "%s - maybe unknown network device or filter by firewall \n", scanhost);
        close(sock);
        exit(0);
    }
    if (strstr(headbuf, "3Com") != NULL) {
        close(sock);
        exit(0);
    }

/* f + r */
    if (ftype == 1) {
        if ((tmp_iprange = fopen(USED_IP_RANGE, "a")) == NULL) {
            fprintf(stderr, "File %s write error", USED_IP_RANGE);
            exit(1);
        } else {
            fprintf(tmp_iprange, "%s\n", scanhost);
            fclose(tmp_iprange);
        }
    } else if (!ruby) {
            if (option == 1)
                ptr = strstr(headbuf, "Allow:");
            else
                ptr = strstr(headbuf, "Server:");
       

    if (ptr == NULL) {
        if (wflags)
            fprintf(htmlreport, "<font size=\"2\"><a href=\"https://%s\">%s</a> - could not get information in header</font><br><br>", 
                scanhost, scanhost);
        else
            fprintf(stdout, "%s - could not get information in header\n", scanhost);
    } else {
        ptr2 = strstr(ptr, "\n");

        *ptr2 = '\0';
        buf_ptr = strdup(ptr);
        *ptr2 = '\n';

        if (wflags && sslflags)
            fprintf(htmlreport, "<br><br><font size=\"2\"><b><a href=\"https://%s\">%s</a> - %s<br></b></font>", 
                scanhost, scanhost, buf_ptr);
        else if (wflags && sslflags == 0 )
            fprintf(htmlreport, "<br><br><font size=\"2\"><b><a href=\"http://%s\">%s</a> - %s<br></b></font>",
                scanhost, scanhost, buf_ptr);
        else
            fprintf(stdout, "%s - %s\n", scanhost, buf_ptr);
        free(buf_ptr);
        }
    } 
/* ruby */  
    else {
    rhost = rb_str_new2(scanhost);
    rb_define_variable("$ari_host", &rhost);
/*	rb_define_variable("$ari_msg", &ari_msg);  */
    ruby_exec();  
/*	fprintf(stdout, "%s \n", StringValuePtr(ari_msg));     */
    }

    if (sslflags)
        ssl_free(ctx, ssl);

    if (wflags)
        fclose(htmlreport);

    if (ruby)
        ruby_finalize();

    close(sock);

}



void
grabheadW(unsigned long host, unsigned int port, unsigned short int type, short int option)
{
    int sock;
    int nb = 0;        
    struct sockaddr_in http;


    char *osfinger;    
    char headbuf[1024];

    char *ptr, *ptr2, *buf_ptr;
    FILE *tmp_iprange;

    SSL_CTX *ctx = NULL;
    SSL *ssl = NULL;
    int err;
    int sread = 0;

    char *rbuf;

    VALUE ari_send;
    VALUE ari_recv;
    VALUE ari_msg;
    VALUE rstr;
    VALUE rhost;
    VALUE ari_host;

    signal(SIGALRM, timegrab);
    alarm(2);

    sock = socket(AF_INET, SOCK_STREAM, 0);
    http.sin_family = AF_INET;
    http.sin_port = htons(port);
    http.sin_addr.s_addr = htonl(host);

#ifdef DEBUG
    printf("Debug %s\n", inet_ntoa(http.sin_addr)); 
#endif

    if (wflags) {
        if ((htmlreport = fopen(html, "a+")) == NULL) {
            fprintf(stderr, "File %s write error\n", html);
            exit(1);
        }
    }

    if (connect(sock, (struct sockaddr *) & http, sizeof(struct sockaddr)) == -1) {
        close(sock);
        exit(0);
    }

    if (sslflags) {
        ctx = ssl_init_ctx();
        ssl = ssl_init_ssl(ctx, sock);
    }

    if (ruby) {
        ruby_init();
        ruby_init_loadpath();

        rb_define_variable("$ari_send", &ari_send);
/*        rb_define_variable("$ari_msg", &ari_msg); */

        rb_load_file(ariruby);
        ruby_exec();
        rbuf = StringValuePtr(ari_send);

        if (strlen(rbuf) == 0) {
            fprintf(stderr, "please set a data in $ari_send variable \n");
            exit(0);
        }

        if ((osfinger = (char *)malloc(strlen(rbuf) + 1)) == NULL)  {
            perror("malloc failed");
            close(sock);
            exit(0);
        }
        strcpy(osfinger, rbuf);
    }
    else if (option) {
        if ((osfinger = (char *)malloc(strlen(SEND_OPTION) + 1)) == NULL) {
            perror("malloc failed");
            close(sock);
            exit(0);
        }
        strcpy(osfinger, SEND_OPTION);
	}
    else {
        if ((osfinger = (char *)malloc(strlen(SEND_URI) + 1)) == NULL) {
            perror("malloc failed");
            close(sock);
            exit(0);
        }
        strcpy(osfinger, SEND_URI);
    }

   if (sslflags) {
        err = SSL_write(ssl, osfinger, strlen(osfinger));
        if (err <= 0) {
        close(sock);
        fprintf(stdout, "%s - ssl write time out \n", inet_ntoa(http.sin_addr));
        exit(0);
        }

    ssl_error(ssl, err, inet_ntoa(http.sin_addr)); 
#ifdef DEBUG
    printf("error = %d\n", err); 
#endif

     sread = ssl_read_timeout(sock, ssl, headbuf, sizeof(headbuf) - 1, CONNECT_TIMEOUT);

    if (ruby) {
        rstr = rb_str_new2(headbuf);
        rb_load_file(ariruby);
        rb_define_variable("$ari_recv", &rstr);
    }

        if (sread <= 0 ) {
            close(sock);

        if (wflags)
            fprintf(htmlreport, "<font size=\"2\"><a href=\"https://%s\">%s</a> - ssl read time out<br></font>",
                inet_ntoa(http.sin_addr), inet_ntoa(http.sin_addr));
        else
            fprintf(stdout, "%s - ssl read time out \n", inet_ntoa(http.sin_addr));
        exit(0);
        }
#ifdef DEBUG
    printf("sread  = %d, buf size= %d \n", sread, strlen(headbuf)); 
#endif
    }
    else {
    send(sock, osfinger, strlen(osfinger), 0);
    nb = recvtimeout(sock, headbuf, sizeof(headbuf), CONNECT_TIMEOUT);
        if (ruby) {
            rstr = rb_str_new2(headbuf);
            rb_load_file(ariruby);
	        rb_define_variable("$ari_recv", &rstr);
/*			rhost = rb_str_new2(inet_ntoa(http.sin_addr));
    		rb_define_variable("$ari_host", &rhost); */
        }
    }

    if (osfinger != NULL)
        free(osfinger);

    if (nb == -1 || nb == -2) { 
#if !(defined(__OpenBSD__) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__APPLE__))
    alarm(0);
#endif
    close(sock);
    exit(0);
    }

    if (strstr(headbuf, "Server Error") != NULL) {
#if !(defined(__OpenBSD__) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__APPLE__))
    alarm(0);
#endif
    close(sock);
    exit(0);
    }

    if (type == 1) {
        if ((tmp_iprange = fopen(USED_IP_RANGE, "a")) == NULL) {
            fprintf(stderr, "File %s write error", USED_IP_RANGE);
            exit(1);
        } else {
            fprintf(tmp_iprange, "%s\n", inet_ntoa(http.sin_addr));
            fclose(tmp_iprange);
        }
    } else if (!ruby) {
            if (option == 1)
                ptr = strstr(headbuf, "Allow:");
            else
                ptr = strstr(headbuf, "Server:");

    if (ptr == NULL) {
        if (wflags)
            fprintf(htmlreport, "<br><font size=\"2\"><a href=\"https://%s\">%s</a> - could not get information in header</font><br>", 
                inet_ntoa(http.sin_addr), inet_ntoa(http.sin_addr));
        else
            fprintf(stdout, "%s - could not get information in header\n", inet_ntoa(http.sin_addr));
    } else {
        ptr2 = strstr(ptr, "\n");

        *ptr2 = '\0';
        buf_ptr = strdup(ptr);
        *ptr2 = '\n';

        if (wflags && sslflags)
            fprintf(htmlreport, "<br><font size=\"2\"><b><a href=\"https://%s\">%s</a> - %s</b><br>", 
                inet_ntoa(http.sin_addr), inet_ntoa(http.sin_addr), buf_ptr);
        else if (wflags && sslflags == 0)
            fprintf(htmlreport, "<br><font size=\"2\"><b><a href=\"http://%s\">%s</a> - %s</b><br>",
                inet_ntoa(http.sin_addr), inet_ntoa(http.sin_addr), buf_ptr);
        else
                fprintf(stdout, "%s - %s\n", inet_ntoa(http.sin_addr), buf_ptr);

            free(buf_ptr);
        }
    }
    else {
    rhost = rb_str_new2(inet_ntoa(http.sin_addr));
    rb_define_variable("$ari_host", &rhost); 
/*  fprintf(stdout, "ruby %s - ", scanhost); */
    ruby_exec();
/*    fprintf(stdout, "%s \n", StringValuePtr(ari_msg)); */

    }   

    if (sslflags)
        ssl_free(ctx, ssl);

    if (wflags)
        fclose(htmlreport);

    if (ruby)
        ruby_finalize();

    close(sock);
}
